/**
  *  \file client/tiles/shipequipment.cc
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipequipment.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "game/spec.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "game/config.h"
#include "ui/widgets.h"
#include "game/cmds.h"
#include "game/global.h"
#include "ui/util.h"
#include "client/remote.h"
#include "game/host.h"

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 6*font_heights[FONT_NORMAL]+3), type),
      button_r(0),
      button_g(0),
      button_s(0)
{
    setLayout(RightJustified, 0, false);

    /* Some buttons: */
    int x = 0;
    if (host.isPHost() && config.CPEnableRemote()) {
        /* Remote control */
        button_r = &h.add(new UIButton("R", 'r'));
        addControl(GfxRect(x, 0, 20, 20), *button_r);
        x += 25;
    }
    if (host.isPHost() && config.CPEnableGive()) {
        /* Give ship */
        button_g = &h.add(new UIButton("G", 'g'));
        addControl(GfxRect(x, 0, 20, 20), *button_g);
        x += 25;
    }

    /* Ship spec sheet */
    button_s = &h.add(new UIButton("S", 's'));
    addControl(GfxRect(x, 0, 20, 20), *button_s);
}

WShipEquipmentTile::~WShipEquipmentTile()
{ }

void
WShipEquipmentTile::drawShipEquipmentTile(UIAbstractTableWriter& tw)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));

    tw.newLColumn();

    const GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    const GShip& ship = *s;
    const GHull& hull = ship.getHull();

    /* engines */
    {
        string_t en = getEngine(ship.getEngine()).getName();
        if (ship.canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (hull.getNumEngines() > 1)
            en = format("%d x %s", hull.getNumEngines(), en);
        tw.writeVariableCell(UIColor::tc_Green, en);
    }

    /* primary */
    if (ship.getBeamCount() && ship.getBeamType()) {
        string_t bm = getBeam(ship.getBeamType()).getName();
        if (ship.getBeamCount() != 1)
            bm = format("%d x %s", ship.getBeamCount(), bm);
        tw.writeVariableCell(UIColor::tc_Green, bm);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* secondary */
    if (ship.getTorpType() && ship.getTorpLauncherCount()) {
        string_t tl = format(_("%d x %s launcher%!1{s%}"),
                             ship.getTorpLauncherCount(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d torpedo%!1{es%} aboard"), ship.getAmmo()))
                       : string_t(_("no torpedoes aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else if (ship.getBayCount()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getBayCount());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d fighter%!1{s%} ready"), ship.getAmmo()))
                       : string_t(_("no fighters aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* damage */
    if (ship.getDamage()) {
        int i = ship.getSupplies() / 5;
        int j = ship.getDamage() - i;
        UISkin::Color c = UIColor::tc_Red;
        if (j <= 0) {
            j = 0;
            c = UIColor::tc_Yellow;
        }
        if (i)
            tw.writeVariableCell(c, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
        else
            tw.writeVariableCell(c, format("%d%%", ship.getDamage()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* crew */
    if (ship.getCrew() != hull.getCrew()) {
        tw.writeVariableCell(ship.getCrew() > hull.getCrew() ? UIColor::tc_Yellow : UIColor::tc_Red,
                             format(_("%d of %d"), ship.getCrew(), hull.getCrew()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, numToString(ship.getCrew()));
    }
    tw.finish();
}

void
WShipEquipmentTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_Fill);
    drawShipEquipmentTile(tw);

    /* Some frames: */
    GShip* sh = dynamic_cast<GShip*>(getObject());
    if (sh != 0) {
        GCommandContainer& con = getDisplayedTurn().getCommands(sh->getOwner());
        if (button_g != 0) {
            /* "give" button */
            drawColorFrame(can, button_g->getExtent(),
                           (con.getCommand(GCommand::phc_GiveShip, sh->getId()) != 0
                            ? cf_Yellow
                            : cf_None),
                           true);
        }
        if (button_r != 0) {
            /* "remote" button */
            drawColorFrame(can, button_r->getExtent(),
                           getRemoteControlColor(getNewRemoteControlState(*sh, con)),
                           true);
        }
        if (button_s != 0) {
            /* "spec" button */
            drawColorFrame(can, button_s->getExtent(),
                           sh->hasAnyShipSpecificSpecials() ? cf_Green : cf_None,
                           true);
        }
    } else {
        /* No ship selected */
        if (button_g != 0)
            drawColorFrame(can, button_g->getExtent(), cf_None, true);
        if (button_r != 0)
            drawColorFrame(can, button_r->getExtent(), cf_None, true);
        if (button_s != 0)
            drawColorFrame(can, button_r->getExtent(), cf_None, true);
    }
}
