/**
  *  \file game/save-dir.cc
  *  \brief Saving Game Data -- Normal unpacked directories
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/checksum.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "game/spec.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/struct.h"
#include "util/misc.h"
#include "cpluslib/misc.h"
#include "game/greg.h"
#include "game/racename.h"

/** Save message outbox, file format 3.5.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
static void
saveOutbox35(GOutbox& outbox, Directory& dir, int player)
{
    Ptr<Stream> f = dir.openFile(format("mess35%d.dat", player), Stream::C_CREATE);
    char header_buffer[19];
    zeroFill(header_buffer);
    storeUint16(header_buffer, outbox.getCount());
    if (!outbox.getCount()) {
        f->writeT(header_buffer, 2);
    } else {
        f->writeT(header_buffer, 19);
        for (int i = 0; i < outbox.getCount(); ++i) {
            /* FIXME: Winplan has a limit of 600 characters. PCC 1.x allows
               to exceed that limit if Winplan is not in use. Right now,
               we enforce the limit. */
            string_t text = outbox.getRawText(i);
            GPlayerSet recv = outbox.getReceiverMask(i);
            encodeMessage(text);
            for (string_t::size_type n = text.length(); n > 0; --n)
                if (text[n-1] == 26)
                    text.insert(n, 1, 23);

            char message_buffer[616];
            message_buffer[0] = 0;     // filler
            message_buffer[1] = '1';   // valid flag
            for (int p = 1; p <= 12; ++p)
                message_buffer[1+p] = (recv.contains(p)) ? '1' : '0';
            storeUint16(message_buffer + 14, 600);

            std::fill_n(message_buffer + 16, 600, char(32+13));
            copy_n(text.begin(), std::min(std::size_t(text.length()), std::size_t(600)), message_buffer + 16);
            f->writeT(message_buffer, 616);
        }
    }
}

/** Save message outbox, file format 3.0.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
static void
saveOutbox(GOutbox& outbox, Directory& dir, int player)
{
    /* count messages */
    int nmessages = 0;
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p)
            if (recv.contains(p))
                ++nmessages;
    }

    // FIXME: some kind of overflow check here?

    /* Allocate buffer for message headers. We leave room for at least
       50 messages after those we wrote, but at least 150 in total. 
       This is to help programs that don't rewrite the message file
       from scratch. PLANETS.EXE has a fixed limit of 50 here; VPA
       seems to relocate messages dynamically when the directory grows
       too large. */
    int nslots = nmessages + 50;
    if (nslots < 150)
        nslots = 150;

    int msgh_size = nslots * TOutgoingMessageHeader::size + 2;
    Array<char> buffer(msgh_size);
    buffer[0U] = buffer[1U] = 0;

    /* write file */
    Ptr<Stream> f = dir.openFile(format("mess%d.dat", player), Stream::C_CREATE);
    f->writeT(buffer.ptr(), msgh_size);

    uint32 output_index = msgh_size + 1;  // +1 for BASIC
    int    header_index = 2;
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (recv.contains(p)) {
                /* send one message */
                string_t text_to_send = outbox.getSendPrefix(i, p) + outbox.getRawText(i);
                encodeMessage(text_to_send);

                TOutgoingMessageHeader h;
                h.address = output_index;
                h.length  = text_to_send.length();
                h.from    = outbox.getSender();
                h.to      = p;
                storeStructure(buffer.ptr() + header_index, h);
                f->writeT(text_to_send.data(), text_to_send.size());
                header_index += TOutgoingMessageHeader::size;
                output_index += text_to_send.length();
            }
        }
    }

    ASSERT(header_index <= msgh_size);
    storeUint16(buffer.ptr(), nmessages);
    f->seek(0);
    f->writeT(buffer, msgh_size);
}

/** Send commands. This appends commands to be sent through the message file
    to the specified GOutbox, and stores the others into the cmdX.txt file
    in the game directory. You must then save he GOutbox.
    \param outbox [in/out] Command messages appended here
    \param cms    [in] Commands
    \param dir    [in] Game directory in which cmdX.txt is created
    \param gen    [in] Game directory control file */
static void
sendCommands(GOutbox& outbox, const GCommandContainer& cms, Directory& dir, const GGen& gen)
{
    Ptr<Stream> cmdfile;
    string_t accum;
    for (GCommandContainer::iterator i = cms.begin(); i != cms.end(); ++i) {
        string_t text = (*i)->getCommandText();
        if (text.length()) {
            if (text[0] != '$') {
                // Send through message file
                if (accum.length() + text.length() > 500) {
                    outbox.addMessage(accum, GPlayerSet(outbox.getSender()));
                    accum.clear();
                }
                if (accum.length())
                    accum += '\n';
                accum += text;
            } else {
                // Send through command file
                if (!cmdfile) {
                    // open file first, and write headers
                    cmdfile = dir.openFile(format("cmd%d.txt", gen.getPlayerId()), Stream::C_CREATE | Stream::C_WRITE);

                    string_t line =  _("# Additional commands");
                    line += Platform::LINE_SEPARATOR;
                    line += "$time ";
                    line += gen.getTimestamp().getTimestampAsString();
                    line += Platform::LINE_SEPARATOR;
                    cmdfile->writeT(line.data(), line.size());
                }
                text += Platform::LINE_SEPARATOR;
                cmdfile->writeT(text.data(), text.size());
            }
        }
    }
    if (accum.length())
        outbox.addMessage(accum, GPlayerSet(outbox.getSender()));

    if (!cmdfile) {
        // Nothing written to command file, so erase it
        dir.eraseNT(format("cmd%d.txt", gen.getPlayerId()));
    }
}

/** Save game directory.
    \param trn         [in] Game turn object to save. This object may be temporarily
                       modified, hence it is not const, but you'll get it back in
                       the same state you gave it to me.
    \param dir         [in] Directory to save in
    \param player_mask [in] Set of players to save */
void
saveDirectory(GGameTurn& trn, Directory& dir, const GPlayerSet player_mask)
{
    bool have_control = false;
    int  control_id   = 0;

    loadFizz();
    for (int pid = 1; pid <= NUM_PLAYERS; ++pid) {
        if (!player_mask.contains(pid))
            continue;

        if (!have_control) {
            // FIXME: this assumes dir == global game directory
            control_id = loadControl(pid);
            have_control = true;
        }

        console.write(LOG_INFO, format(_("Writing %s data..."), player_racenames.getAdjName(pid)));

        {
            Ptr<Stream> s;

            GGen& g = trn.getGen(pid);
            uint32 sig_check = getChecksum(g.getSig1(), 10) + getChecksum(g.getSig2(), 10);

            uint32 ship_chk = 0, planet_chk = 0, base_chk = 0;

            s = dir.openFile(format("ship%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveShips(trn.getCurrentUniverse(), *s, GPlayerSet(pid), &ship_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            s = dir.openFile(format("pdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            savePlanets(trn.getCurrentUniverse(), *s, GPlayerSet(pid), &planet_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            s = dir.openFile(format("bdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveBases(trn.getCurrentUniverse(), *s, GPlayerSet(pid), &base_chk, save_ChecksumCount);
            s->writeT(trn.getGen(pid).getSig2(), 10);

            /* messages and commands. We add the commands to the message
               box, save that, and remove them again. */
            GOutbox& out = trn.getOutbox(pid);
            int marker = out.getCount();
            sendCommands(out, trn.getCommands(pid), dir, g);
            if (out.getFormat() == GOutbox::fWinplan)
                saveOutbox35(out, dir, pid);
            else
                saveOutbox(out, dir, pid);
            out.deleteMessagesAfter(marker);

            /* compute DIS checksums */
            ship_chk += computeShipChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            planet_chk += computePlanetChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            base_chk += computeBaseChecksum(trn.getPreviousUniverse(), GPlayerSet(pid), save_ChecksumCount);
            
            g.setChecksum(ShipSection,       ship_chk   + sig_check);
            g.setChecksum(PlanetSection,     planet_chk + sig_check);
            g.setChecksum(BaseSection,       base_chk   + sig_check);
            setFizzEntry(ShipSection,   pid, ship_chk   + sig_check);
            setFizzEntry(PlanetSection, pid, planet_chk + sig_check);
            setFizzEntry(BaseSection,   pid, base_chk   + sig_check);
            s = dir.openFile(format("gen%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            storeStructure(*s, g.getGenRecord());

            updateGameRegistry(g.getTimestamp());
        }

        if (control_id > 0) {
            // we are playing DOS planets
            saveControl(control_id);
            have_control = false;
        }
    }

    if (have_control)
        saveControl(control_id);
    saveFizz();
}
