/**
  *  \file int/bytecode.h
  *  \brief Interpreter: Bytecode Object
  *
  *  \todo Can we make referred BCOs literals? pro: this would save a special
  *  case and an addressing mode. con: we would need to be able to serialize
  *  BCOs in IntDataSegment.
  */
#ifndef PCC_V2_INT_BYTECODE_H
#define PCC_V2_INT_BYTECODE_H

#include <vector>
#include "cpluslib/types.h"
#include "int/data.h"
#include "int/opcode.h"
#include "int/names.h"
#include "cpluslib/smartptr.h"

class IntBytecodeObject;
class IntCompilationContext;

/** Reference to an IntBytecodeObject. */
typedef Ptr<IntBytecodeObject> IntBCORef;

/** Bytecode object (BCO). Bytecode objects contain code for execution.
    It contains the following elements:
    - actual code
    - a list of literals referred to by the code (only small integer literals
      can be encoded directly in the code, others are loaded from this table).
    - a list of names (symbols) referred to by the code.
    - a list of other BCOs referred to by the code. This means only subroutines
      <em>defined</em> by this BCO, not <em>called</em> subroutines.

    For the benefit of code generation, the BCO can also generate symbolic
    labels.

    BCOs use reference counting. Each IntExecutionFrame executing a BCO refers
    to it, as well as the symbol table entry holding it, as well as another
    BCO which happened to define it. Circular references are impossible. For
    example, assume we are executing "foo" and are at the line containing "bar":
    <pre>
      Sub foo
        Sub bar
          Print "hi"
        EndSub
        bar
      EndSub
    </pre>
    The BCO "foo" will be referenced from the symbol table and the executing
    frame. The BCO "bar" will be referenced from the symbol table (it has just
    been defined), from the executing frame, and from the "foo" BCO.

    \todo BCOs should manage parameter counts (min, max, [array-flag]), and
    predeclared identifiers. */
class IntBytecodeObject {
 public:
    typedef std::vector<IntOpcode>::size_type pc_t;
    typedef uint16_t label_t;

    IntBytecodeObject();
    ~IntBytecodeObject();

    label_t  makeLabel();
    void     addInstruction(IntOpcode::Major major, uint8_t minor, uint16_t arg);
    void     addVariableReferenceInstruction(IntOpcode::Major major, const string_t& name, const IntCompilationContext& cc);
    void     addLabel(label_t label);
    void     addJump(uint8_t flags, label_t label);
    void     addPushLiteral(IntValue* literal);
    uint16_t addName(string_t name);
    void     relocate();

    pc_t     getNumInstructions() const;
    pc_t     getJumpTarget(uint8_t minor, uint16_t arg);
    IntOpcode&       operator()(pc_t index);
    const IntOpcode& operator()(pc_t index) const;

    IntValue* getLiteral(uint16_t index);
    string_t  getName(uint16_t index);
    IntBCORef getSubroutine(uint16_t index);

 private:
    IntDataSegment         data;   ///< Literals referenced in bytecode.
    IntVariableNames       names;  ///< Names referenced in bytecode.
    std::vector<IntOpcode> code;   ///< Actual code.
    label_t                num_labels;
    std::vector<IntBCORef> subrs;  ///< Inline subroutines referenced in code.
    IntVariableNames       local_names;
};

/**************************** Inline Functions ***************************/

/** Get number of instructions in this bytecode object. */
inline IntBytecodeObject::pc_t
IntBytecodeObject::getNumInstructions() const
{
    return code.size();
}

/** Read/write access to instruction by address. */
inline IntOpcode&
IntBytecodeObject::operator()(pc_t index)
{
    return code[index];
}

/** Read access to instruction by address. */
inline const IntOpcode&
IntBytecodeObject::operator()(pc_t index) const
{
    return code[index];
}

#endif
