/**
  *  \file int/context.h
  *  \brief Interpreter: Contexts
  */
#ifndef PCC_V2_INT_CONTEXT_H
#define PCC_V2_INT_CONTEXT_H

#include "cpluslib/types.h"
#include "cpluslib/object.h"
#include "cpluslib/string.h"

class IntExecutionFrame;
class IntValue;

/** Context for name lookup. A context provides a means for looking up and
    dealing with local names, and possibly iteration through objects. */
class IntContext : public Object {
 public:
    /** Look up a symbol by its name.
        \return index for symbol (>= 0), or -1 for failure. */
    virtual int32_t lookup(string_t name) = 0;

    /** Set value by its index. The parameter is owned by the caller. */
    virtual void set(int32_t index, IntValue* value) = 0;

    /** Get value by its index. The returned value must be newly allocated,
        caller assumes responsibility. */
    virtual IntValue* get(int32_t index) = 0;

    /** Advance to next object. Return true on success, false on failure. */
    virtual bool next() = 0;

    /** Clone this context. */
    virtual IntContext* clone() = 0;
};

/** Single context. This is a context that tells its user that it does not
    support iteration. */
class IntSingleContext : public IntContext {
 public:
    bool next();
};

/** Local variables for an execution frame. Provides access to the local variables
    of an executing stack frame. */
class IntExecutionFrameContext : public IntSingleContext {
    IntExecutionFrame& frame;
 public:
    IntExecutionFrameContext(IntExecutionFrame& frame);
    ~IntExecutionFrameContext();

    int32_t   lookup(string_t name);
    void      set(int32_t index, IntValue* name);
    IntValue* get(int32_t index);
    IntExecutionFrameContext* clone();
};

#endif
