/**
  *  \file int/globalif.cc
  *  \brief Interpreter Game Interface: Global Functions and Properties
  */

#include <memory>
#include <cmath>
#include "int/if/globalif.h"
#include "game/coord.h"
#include "int/data.h"
#include "int/value.h"
#include "int/compound.h"
#include "int/context.h"
#include "int/interror.h"
#include "int/if/ifutil.h"
#include "game/struct.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "game/config.h"
#include "game/team.h"
#include "version.h"
#include "game/host.h"
#include "cpluslib/misc.h"
#include "int/intglobal.h"
#include "game/fcode.h"
#include "int/if/playerif.h"

namespace {
    enum { GlobalProperty, MyPlayerProperty };
}

static const TIntMapping global_mapping[] = {
    { "MY.BASES",              iplScoreBases,        MyPlayerProperty },
    { "MY.PLANETS",            iplScorePlanets,      MyPlayerProperty },
    { "MY.RACE",               iplShortName,         MyPlayerProperty },
    { "MY.RACE$",              iplId,                MyPlayerProperty },
    { "MY.RACE.ADJ",           iplAdjName,           MyPlayerProperty },
    { "MY.RACE.FULL",          iplFullName,          MyPlayerProperty },
    { "MY.RACE.ID",            iplRaceId,            MyPlayerProperty },
    { "MY.RACE.MISSION",       iplMission,           MyPlayerProperty },
    { "MY.SCORE",              iplScore,             MyPlayerProperty },
    { "MY.SHIPS",              iplScoreShips,        MyPlayerProperty },
    { "MY.SHIPS.CAPITAL",      iplScoreCapital,      MyPlayerProperty },
    { "MY.SHIPS.FREIGHTERS",   iplScoreFreighters,   MyPlayerProperty },
    { "MY.TEAM",               iplTeam,              MyPlayerProperty },
    { "SYSTEM.HOST",           igpSystemHost,        GlobalProperty },
    { "SYSTEM.HOST$",          igpSystemHostCode,    GlobalProperty },
    { "SYSTEM.HOSTVERSION",    igpSystemHostVersion, GlobalProperty },
    { "SYSTEM.PROGRAM",        igpSystemProgram,     GlobalProperty },
    { "SYSTEM.VERSION",        igpSystemVersion,     GlobalProperty },
    { "SYSTEM.VERSION$",       igpSystemVersionCode, GlobalProperty },
};

/**************************** IntGlobalContext ***************************/

IntGlobalContext::IntGlobalContext()
{ }

IntGlobalContext::~IntGlobalContext()
{ }

int32_t
IntGlobalContext::lookup(string_t name)
{
    static const uint32_t nproperties = countof(global_mapping);

    if (name.compare(0, 7, "GLOBAL.", 7) == 0)
        name.erase(0, 7);

    int32_t result = global_property_names.lookup(name);
    if (result >= 0)
        return result + nproperties;

    return lookupName(name, global_mapping, countof(global_mapping));
}

void
IntGlobalContext::set(int32_t index, IntValue* value)
{
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        global_values.set(index - nproperties, value);
    } else {
        throw IntError::notAssignable();
    }
}

IntValue*
IntGlobalContext::get(int32_t index)
{
    /* Check user variables */
    static const int32_t nproperties = countof(global_mapping);
    if (index >= nproperties) {
        return IntValue::cloneOf(global_values[index - nproperties]);
    }

    /* Check own properties */
    switch (global_mapping[index].domain) {
     case GlobalProperty:
        return getGlobalProperty(IntGlobalProperty(global_mapping[index].index));
     case MyPlayerProperty:
        return getPlayerProperty(getPlayerId(), IntPlayerProperty(global_mapping[index].index));
    }
    return 0;
// Always valid:
//   System.Egg: Int {%Egg}{Egg:=%%}
//   System.Sim: Bool {%Ord(InSim)}
//   System.GameType: String {IF Reg THEN %'Shareware' ELSE %'Registered'}
//   System.GameType$: Bool {%Ord(Reg)}
//   System.GameDirectory: String {%GameDir}
//   System.RootDirectory: String {%HostDir}
//   System.GUI: Bool {%Ord(Gr)}
//   System.Local: String {
//     CASE CurrentGameType OF
//       1:%'DOS';
//       2:%'Windows';
//     ELSE %'PCC';
//     END}
//   System.RegStr1: String {%GetStr(RegStr[1], 25)}
//   System.RegStr2: String {%GetStr(RegStr[2], 25)}
//   System.Remote: String {IF Odd(GameInfo.Remote) THEN %'Windows' ELSE %'DOS'}
//   System.Err: String {%FPError_Last}
//   System.RandomSeed: Int {%ScriptRandSeed}{ScriptRandSeed:=%%}
//   Turn: Int {%Gen.TurnNr}
//   Turn.Date: String {%GetStr(Gen.Timestamp, 10)}
//   Turn.IsNew: Bool {%Ord(IsNewTurn)}
//   Turn.Time: String {%GetStr(Gen.Timestamp[11], 8)}

// valid when player loaded:
//   My.InMsgs: Int {%MsgInCnt}
//   My.OutMsgs: Int {%MsgOutCnt}
//   My.PBPs: Int {%PBPs[PlayerId]}
//   My.Team: Int {%TeamData.PlayerTeams[PlayerId]}
//   My.VCRs: Int {%VCRvisible}
//   Ships.Total: Int, Ships.Capital: Int, Ships.Freighters: Int {
//     VAR cap, fre, j:INTEGER;
//     cap := 0; fre := 0;
//     FOR j:=1 TO 11 DO BEGIN
//       Inc(cap,Gen.Scores[j].Capital);
//       Inc(fre,Gen.Scores[j].Freighters);
//     END;
//     IF $Ships.Total THEN %(cap+fre) ELSE
//     IF $Ships.Capital THEN %cap ELSE %fre;}

//   UI.X: Int {VAR i:INTEGER;
//     IF current_control_screen<>NIL THEN BEGIN
//       i := current_control_screen^.scanner^.chart^.scanx;
//       IF i>0 THEN %i;
//     END ELSE %ChartX0}
//   UI.Y: Int {VAR i:INTEGER;
//     IF current_control_screen<>NIL THEN BEGIN
//       i := current_control_screen^.scanner^.chart^.scany;
//       IF i>0 THEN %i;
//     END ELSE %ChartY0}
//   UI.Screen: Int {%current_screen_id}
//   Chart.X: Int {%ChartX0}{StarchartLabelsChanged:=TRUE; ChartX0 := %%}
//   Chart.Y: Int {%ChartY0}{StarchartLabelsChanged:=TRUE; ChartY0 := %%}

//   Selection.Layer: Int {%CurrentSelection}{
//     IF (%%<0) OR (%%>=8) THEN FAIL ELSE
//     IF INTEGER(%%)<>CurrentSelection THEN BEGIN
//       StoreSelectionLayer(CurrentSelection);
//       CurrentSelection := %%;
//       GetSelectionLayer(CurrentSelection, TRUE);
//     END;
//   }
}

IntContext*
IntGlobalContext::clone()
{
    return new IntGlobalContext();
}


/** Get global property.
    \param igp Global property to query. */
IntValue*
getGlobalProperty(IntGlobalProperty igp)
{
    switch (igp) {
     case igpSystemProgram:
        return makeStringValue("PCC");
     case igpSystemVersion:
        return makeStringValue(PCC2_VERSION);
     case igpSystemVersionCode:
        return makeIntValue(PCC2_VERSION_CODE);
     case igpSystemHost:
        if (host.isPHost())
            return makeStringValue("PHost");
        else if (host.isSRace())
            return makeStringValue("SRace");
        else if (host.isTHost())
            return makeStringValue("Host");
        else
            return 0;
     case igpSystemHostCode:
        if (host.isPHost())
            return makeIntValue(2);
        else if (host.isSRace())
            return makeIntValue(1);
        else if (host.isTHost())
            return makeIntValue(0);
        else
            return 0;
     case igpSystemHostVersion:
        return makeIntValue(host.host_version);
    }
    return 0;
}

/******************************* Functions *******************************/

/** Implementation of the "Distance" function. */
IntValue*
IFDistanceGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    GPoint args[2];

    for (int i = 0; i < 2; ++i) {
        /* Do we have an argument? */
        if (nargs == 0)
            throw IntError("Too few arguments for \"Distance\"");
        IntValue* theValue = dseg[index++];
        --nargs;

        /* What is it? */
        if (theValue == 0) {
            /* Null. Result is null. */
            return 0;
        } else if (IntContextValue* cv = dynamic_cast<IntContextValue*>(theValue)) {
            /* Context. Must have LOC.X and LOC.Y properties. */
            std::auto_ptr<IntContext> c(cv->makeContext());
            int32_t x = c->lookup("LOC.X");
            int32_t y = c->lookup("LOC.Y");
            if (x < 0 || y < 0)
                throw IntError("Operand doesn't have a position");

            /* Fetch values */
            std::auto_ptr<IntValue> vx(c->get(x));
            std::auto_ptr<IntValue> vy(c->get(y));
            if (!vx.get() || !vy.get())
                return 0;

            /* Must be integers */
            IntIntValue* vix = dynamic_cast<IntIntValue*>(vx.get());
            IntIntValue* viy = dynamic_cast<IntIntValue*>(vy.get());
            if (!vix || !viy)
                throw IntError::typeError(IntError::tyInteger);

            args[i].x = vix->getValue();
            args[i].y = viy->getValue();
        } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(theValue)) {
            /* Integer. There must be another integer. */
            if (nargs == 0)
                throw IntError("Too few arguments for \"Distance\"");
            IntValue* otherValue = dseg[index++];
            --nargs;

            /* Null? */
            if (!otherValue)
                return 0;

            /* Integer? */
            IntIntValue* iv2 = dynamic_cast<IntIntValue*>(otherValue);
            if (!iv2)
                throw IntError::typeError(IntError::tyInteger);

            args[i].x = iv->getValue();
            args[i].y = iv2->getValue();
        } else {
            /* Error. */
            throw IntError::typeError();
        }
    }

    if (nargs != 0)
        throw IntError("Too many arguments to \"Distance\"");

    return makeFloatValue(std::sqrt(double(distanceSquared(args[0], args[1]))));
}

/** Implementation of the "ShipName" function. */
IntValue*
IFShipNameGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t sid;

    checkArgumentCount(nargs, 1);
    if (!checkIntArg(sid, dseg[index]))
        return 0;

    if (sid <= 0 || sid > NUM_SHIPS)
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sid).getName());
}

/** Implementation of the "PlanetName" function. */
IntValue*
IFPlanetNameGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t sid;

    checkArgumentCount(nargs, 1);
    if (!checkIntArg(sid, dseg[index]))
        return 0;

    if (sid <= 0 || sid > NUM_PLANETS)
        return 0;

    return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sid).getName());
}

/** Implementation of the "Cfg" function. */
IntValue*
IFCfgGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    checkArgumentCount(nargs, 1, 2);

    /* Config key */
    IntValue* v1 = dseg[index++];
    if (v1 == 0)
        return 0;
    IntStringValue* sv1 = dynamic_cast<IntStringValue*>(v1);   /* FIXME: 1.x stringifies here */
    if (sv1 == 0)
        throw IntError::typeError(IntError::tyString);
    ConfigOption* opt = config.getConfigOption(sv1->getValue());
    if (opt == 0)
        throw IntError("Invalid first argument to \"Cfg\"");

    /* Player number */
    int32_t player = 0;
    if (nargs > 1)
        if (!checkIntArg(player, dseg[index++], 1, NUM_PLAYERS))
            throw IntError("Invalid second argument to \"Cfg\"");

    /* Figure out option */
    if (ConfigAlias* a = dynamic_cast<ConfigAlias*>(opt))
        opt = &a->getForwardedOption();

    /* FIXME: ugly type switch. Can we do better? */
    if (ConfigIntArrayBaseOption* bopt = dynamic_cast<ConfigIntArrayBaseOption*>(opt)) {
        /* ints, optional player */
        if (!player) {
            /* Possible limits are
                 2    NewNativesPopulationRange
                 4    WraparoundRectangle
                 8    MeteorShowerOreRanges
                 9    NewNativesRaceRate
                 10   ConfigExpOption, e.g. EModBayRechargeRate
                 11   ConfigStdOption, e.g. RaceMiningRate */
            if (bopt->getLimit() == 11)
                player = getPlayerId();
            else
                throw IntError("Too few arguments");
        }
        if (player > bopt->getLimit())
            throw IntError::rangeError();
        return makeIntValue(bopt->getArray()[player-1]);
    } else if (ConfigIntOption* intopt = dynamic_cast<ConfigIntOption*>(opt)) {
        /* single int, no player. Example: NumShips */
        if (player)
            throw IntError("Too many arguments");
        return makeIntValue((*intopt)());
    } else if (ConfigCostOption* costopt = dynamic_cast<ConfigCostOption*>(opt)) {
        /* 11 costs. Example: StarbaseCost */
        if (!player)
            player = getPlayerId();
        return makeStringValue((*costopt)(player).toCargospec());        
    } else if (ConfigStringOption* stropt = dynamic_cast<ConfigStringOption*>(opt)) {
        /* single string. FIXME: PCC 1.x splits ExperienceLevelNames */
        if (player)
            throw IntError("Too many arguments");
        return makeStringValue((*stropt)());
    } else {
        /* what? */
        return 0;
    }
}

/** Implementation of the "IsSpecialFCode" function. */
IntValue*
IFIsSpecialFCodeGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    /* Check arguments */
    checkArgumentCount(nargs, 1);
    if (!dseg[index])
        return 0;

    /* Function implementation */
    string_t fc = dseg[index]->toString(false);
    return makeBoolValue(fc.size() <= 3 && GFCode::isSpecial(fc, true));
}

/** Implementation of the "Truehull" function. */
IntValue*
IFTruehull(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t slot, player;

    /* Check arguments */
    checkArgumentCount(nargs, 1, 2);
    if (!checkIntArg(slot, dseg[index], 1, NUM_HULLS_PER_PLAYER))
        return 0;

    if (nargs >= 2) {
        if (!checkIntArg(player, dseg[index+1], 1, NUM_PLAYERS))
            return 0;
    } else {
        player = getPlayerId();
    }

    return makeIntValue(getTruehull(player, slot));
}

/** Implementation of the "RandomFCode" function. */
IntValue*
IFRandomFCode(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    (void) dseg;
    (void) index;
    checkArgumentCount(nargs, 0);
    return makeStringValue(GFCode::generateRandomFCode());
}
