/**
  *  \file int/if/shipif.cc
  *  \brief Interpreter Game Interface: Ships
  */

#include "int/if/shipif.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/objtypes.h"
#include "game/ship.h"
#include "int/if/hullif.h"
#include "int/intglobal.h"
#include "cpluslib/misc.h"
#include "int/interror.h"
#include "game/racename.h"
#include "game/tables.h"
#include "game/mission.h"
#include "game/formula.h"
#include "game/planet.h"
#include "int/if/ifutil.h"
#include "int/if/specif.h"
#include "int/if/playerif.h"

/****************************** IntShipValue *****************************/

/** Implementation of "Ship(nn)".
    \param dseg Data segment containing function parameters
    \param index Index of function parameters
    \param nargs Number of function parameters
    \return value representing "Ship(nn)" */
IntValue*
IFShipGet(IntDataSegment& dseg, uint32_t index, uint32_t nargs)
{
    int32_t id;

    checkArgumentCount(nargs, 1);
    if (!checkIntArg(id, dseg[index], 1, NUM_SHIPS))
        return 0;

    return IntSimpleContextValue::createFromNewContext(new IntShipContext(id));
}

/** Implementation of "ForEach Ship".
    \return Context representing first ship */
IntContext*
IFShipMake()
{
    int id = global_anyship_type.getNextNoWrap(0, false);
    if (id == 0)
        return 0;
    else
        return new IntShipContext(id);
}

/***************************** IntShipContext ****************************/

namespace {
    enum { ShipProperty, HullProperty, ComponentProperty, OwnerProperty, EnemyProperty };
}

static const TIntMapping ship_mapping[] = {
    { "AUX",                       ispAUX,                       ShipProperty },
    { "AUX$",                      ispAUX_,                      ShipProperty },
    { "AUX.AMMO",                  ispAUX_AMMO,                  ShipProperty },
    { "AUX.COUNT",                 ispAUX_COUNT,                 ShipProperty },
    { "AUX.SHORT",                 ispAUX_SHORT,                 ShipProperty },
    { "BEAM",                      ispBEAM,                      ShipProperty },
    { "BEAM$",                     ispBEAM_,                     ShipProperty },
    { "BEAM.COUNT",                ispBEAM_COUNT,                ShipProperty },
    { "BEAM.MAX",                  ihpMaxBeams,                  HullProperty },
    { "BEAM.SHORT",                ispBEAM_SHORT,                ShipProperty },
    { "CARGO.COLONISTS",           ispCARGO_COLONISTS,           ShipProperty },
    { "CARGO.D",                   ispCARGO_D,                   ShipProperty },
    { "CARGO.FREE",                ispCARGO_FREE,                ShipProperty },
    { "CARGO.M",                   ispCARGO_M,                   ShipProperty },
    { "CARGO.MAX",                 ihpMaxCargo,                  HullProperty },
    { "CARGO.MAXFUEL",             ihpMaxFuel,                   HullProperty },
    { "CARGO.MONEY",               ispCARGO_MONEY,               ShipProperty },
    { "CARGO.N",                   ispCARGO_N,                   ShipProperty },
    { "CARGO.STR",                 ispCARGO_STR,                 ShipProperty },
    { "CARGO.SUPPLIES",            ispCARGO_SUPPLIES,            ShipProperty },
    { "CARGO.T",                   ispCARGO_T,                   ShipProperty },
    { "CREW",                      ispCREW,                      ShipProperty },
    { "CREW.NORMAL",               ihpMaxCrew,                   HullProperty },
    { "DAMAGE",                    ispDAMAGE,                    ShipProperty },
    { "ENEMY",                     iplShortName,                 EnemyProperty },
    { "ENEMY$",                    iplId,                        EnemyProperty },
    { "ENEMY.ADJ",                 iplAdjName,                   EnemyProperty },
    { "ENGINE",                    ispENGINE,                    ShipProperty },
    { "ENGINE$",                   ispENGINE_,                   ShipProperty },
    { "ENGINE.COUNT",              ihpNumEngines,                HullProperty },
    { "FCODE",                     ispFCODE,                     ShipProperty },
    { "FIGHTER.BAYS",              ispFIGHTER_BAYS,              ShipProperty },
    { "FIGHTER.COUNT",             ispFIGHTER_COUNT,             ShipProperty },
    { "FLEET",                     ispFLEET,                     ShipProperty },
    { "FLEET$",                    ispFLEET_,                    ShipProperty },
    { "FLEET.NAME",                ispFLEET_NAME,                ShipProperty },
    { "FLEET.STATUS",              ispFLEET_STATUS,              ShipProperty },
    { "HEADING",                   ispHEADING,                   ShipProperty },
    { "HEADING$",                  ispHEADING_,                  ShipProperty },
    { "HULL",                      icpName,                      ComponentProperty },
    { "HULL$",                     icpId,                        ComponentProperty },
    { "HULL.SHORT",                icpNameShort,                 ComponentProperty },
    { "HULL.SPECIAL",              ispHULL_SPECIAL,              ShipProperty },
    { "ID",                        ispID,                        ShipProperty },
    { "LEVEL",                     ispLEVEL,                     ShipProperty },
    { "LOC",                       ispLOC,                       ShipProperty },
    { "LOC.X",                     ispLOC_X,                     ShipProperty },
    { "LOC.Y",                     ispLOC_Y,                     ShipProperty },
    { "MARKED",                    ispMARKED,                    ShipProperty },
    { "MASS",                      ispMASS,                      ShipProperty },
    { "MISSION",                   ispMISSION,                   ShipProperty },
    { "MISSION$",                  ispMISSION_,                  ShipProperty },
    { "MISSION.INTERCEPT",         ispMISSION_INTERCEPT,         ShipProperty },
    { "MISSION.SHORT",             ispMISSION_SHORT,             ShipProperty },
    { "MISSION.TOW",               ispMISSION_TOW,               ShipProperty },
    { "MOVE.ETA",                  ispMOVE_ETA,                  ShipProperty },
    { "MOVE.FUEL",                 ispMOVE_FUEL,                 ShipProperty },
    { "NAME",                      ispNAME,                      ShipProperty },
    { "ORBIT",                     ispORBIT,                     ShipProperty },
    { "ORBIT$",                    ispORBIT_,                    ShipProperty },
    { "OWNER",                     iplShortName,                 OwnerProperty },
    { "OWNER$",                    iplId,                        OwnerProperty },
    { "OWNER.ADJ",                 iplAdjName,                   OwnerProperty },
    { "OWNER.REAL",                ispOWNER_REAL,                ShipProperty },
    { "SPEED",                     ispSPEED,                     ShipProperty },
    { "SPEED$",                    ispSPEED_,                    ShipProperty },
    { "TASK",                      ispTASK,                      ShipProperty },
    { "TECH.HULL",                 icpTech,                      ComponentProperty },
    { "TORP",                      ispTORP,                      ShipProperty },
    { "TORP$",                     ispTORP_,                     ShipProperty },
    { "TORP.COUNT",                ispTORP_COUNT,                ShipProperty },
    { "TORP.LCOUNT",               ispTORP_LCOUNT,               ShipProperty },
    { "TORP.LMAX",                 ihpMaxTorpLaunchers,          HullProperty },
    { "TORP.SHORT",                ispTORP_SHORT,                ShipProperty },
    { "TRANSFER.SHIP",             ispTRANSFER_SHIP,             ShipProperty },
    { "TRANSFER.SHIP.COLONISTS",   ispTRANSFER_SHIP_COLONISTS,   ShipProperty },
    { "TRANSFER.SHIP.D",           ispTRANSFER_SHIP_D,           ShipProperty },
    { "TRANSFER.SHIP.ID",          ispTRANSFER_SHIP_ID,          ShipProperty },
    { "TRANSFER.SHIP.M",           ispTRANSFER_SHIP_M,           ShipProperty },
    { "TRANSFER.SHIP.N",           ispTRANSFER_SHIP_N,           ShipProperty },
    { "TRANSFER.SHIP.NAME",        ispTRANSFER_SHIP_NAME,        ShipProperty },
    { "TRANSFER.SHIP.SUPPLIES",    ispTRANSFER_SHIP_SUPPLIES,    ShipProperty },
    { "TRANSFER.SHIP.T",           ispTRANSFER_SHIP_T,           ShipProperty },
    { "TRANSFER.UNLOAD",           ispTRANSFER_UNLOAD,           ShipProperty },
    { "TRANSFER.UNLOAD.COLONISTS", ispTRANSFER_UNLOAD_COLONISTS, ShipProperty },
    { "TRANSFER.UNLOAD.D",         ispTRANSFER_UNLOAD_D,         ShipProperty },
    { "TRANSFER.UNLOAD.ID",        ispTRANSFER_UNLOAD_ID,        ShipProperty },
    { "TRANSFER.UNLOAD.M",         ispTRANSFER_UNLOAD_M,         ShipProperty },
    { "TRANSFER.UNLOAD.N",         ispTRANSFER_UNLOAD_N,         ShipProperty },
    { "TRANSFER.UNLOAD.NAME",      ispTRANSFER_UNLOAD_NAME,      ShipProperty },
    { "TRANSFER.UNLOAD.SUPPLIES",  ispTRANSFER_UNLOAD_SUPPLIES,  ShipProperty },
    { "TRANSFER.UNLOAD.T",         ispTRANSFER_UNLOAD_T,         ShipProperty },
    { "TYPE",                      ispTYPE,                      ShipProperty },
    { "TYPE.SHORT",                ispTYPE_SHORT,                ShipProperty },
    { "WAYPOINT",                  ispWAYPOINT,                  ShipProperty },
    { "WAYPOINT.DIST",             ispWAYPOINT_DIST,             ShipProperty },
    { "WAYPOINT.DX",               ispWAYPOINT_DX,               ShipProperty },
    { "WAYPOINT.DY",               ispWAYPOINT_DY,               ShipProperty },
    { "WAYPOINT.PLANET",           ispWAYPOINT_PLANET,           ShipProperty },
    { "WAYPOINT.X",                ispWAYPOINT_X,                ShipProperty },
    { "WAYPOINT.Y",                ispWAYPOINT_Y,                ShipProperty },
};


IntShipContext::IntShipContext(int sid)
    : sid(sid)
{ }

IntShipContext::~IntShipContext()
{ }

int32_t
IntShipContext::lookup(string_t name)
{
    static const int32_t nproperties = countof(ship_mapping);

    /* Check prefix */
    if (name.compare(0, 5, "SHIP.", 5) == 0)
        name.erase(0, 5);

    /* Check user-defined properties */
    IntVariableNames::index_t ix = ship_property_names.lookup(name);
    if (ix != ship_property_names.nil)
        return ix + nproperties;

    /* Check predefined properties */
    return lookupName(name, ship_mapping, nproperties);
}

void
IntShipContext::set(int32_t index, IntValue* value)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip& sh = global_anyship_type.getObject(sid);
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            setShipProperty(sh, IntShipProperty(ship_mapping[index].index), value);
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        /* User property */
        getShipProperties(sid).set(index - countof(ship_mapping), value);
    }
}

IntValue*
IntShipContext::get(int32_t index)
{
    if (index < int32_t(countof(ship_mapping))) {
        /* Builtin property */
        GShip& sh = global_anyship_type.getObject(sid);
        switch (ship_mapping[index].domain) {
         case ShipProperty:
            return getShipProperty(sh, IntShipProperty(ship_mapping[index].index));
         case HullProperty:
            if (sh.getHullId() > 0)
                return getHullProperty(sh.getHull(), IntHullProperty(ship_mapping[index].index));
            else
                return 0;
         case ComponentProperty:
            if (sh.getHullId() > 0)
                return getComponentProperty(sh.getHull(), IntComponentProperty(ship_mapping[index].index));
            else
                return 0;
         case OwnerProperty:
            return getPlayerProperty(sh.getOwner(), IntPlayerProperty(ship_mapping[index].index));
         case EnemyProperty:
            if (sh.getPE().isKnown())
                return getPlayerProperty(sh.getPE(), IntPlayerProperty(ship_mapping[index].index));
            else
                return 0;
         default:
            return 0;
        }
    } else {
        /* User property */
        return IntValue::cloneOf(getShipProperties(sid)[index - countof(ship_mapping)]);
    }
}

bool
IntShipContext::next()
{
    int next = global_anyship_type.getNextNoWrap(sid, false);
    if (next != 0) {
        sid = next;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntShipContext::clone()
{
    return new IntShipContext(sid);
}

static const char*
classifyShip(const GShip& sh)
{
    int beams = sh.getBeamCount().isKnown()         ? sh.getBeamCount()         : sh.getHullId() > 0 ? sh.getHull().getMaxBeams()     : -1;
    int tubes = sh.getTorpLauncherCount().isKnown() ? sh.getTorpLauncherCount() : sh.getHullId() > 0 ? sh.getHull().getMaxLaunchers() : -1;
    int bays  = sh.getBayCount().isKnown()          ? sh.getBayCount()          : sh.getHullId() > 0 ? sh.getHull().getNumBays()      : -1;

    if (bays > 0)
        return "Carrier";
    else if (tubes > 0)
        return "Torpedo Ship";
    else if (beams > 0)
        return "Beam Weapons";
    else if (tubes == 0 && beams == 0 && bays == 0)
        return "Freighter";
    else
        return 0;
}

static IntValue*
makeIntFromMaybe(mp16_t value)
{
    if (value.isKnown())
        return makeIntValue(value);
    else
        return 0;
}

IntValue*
getShipProperty(GShip& sh, IntShipProperty isp)
{
    switch (isp) {
     case ispAUX_:
        /* 11 for fighters, torpedo tech if any, otherwise null */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeIntValue(11);
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeIntValue(sh.getTorpType());
        else
            return 0;
     case ispAUX_AMMO:
        /* ammo */
        return makeIntValue(sh.getAmmo());
     case ispAUX_COUNT:
        /* number of aux weapons */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeIntValue(sh.getBayCount());
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0 && sh.getTorpLauncherCount().isKnown())
            return makeIntValue(sh.getTorpLauncherCount());
        else
            return 0;
     case ispAUX_SHORT:
        /* short name of aux weapon */
        return 0;               // FIXME
     case ispAUX:
        /* full name of aux weapon */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0)
            return makeStringValue("Fighters");
        else if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispBEAM_:
        /* beam type number */
        return makeIntFromMaybe(sh.getBeamType());
     case ispBEAM_COUNT:
        /* beam count */
        return makeIntFromMaybe(sh.getBeamCount());
     case ispBEAM_SHORT:
        /* short name of beam */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getShortName());
        else
            return 0;
     case ispBEAM:
        /* beam type */
        if (sh.getBeamType().isKnown() && sh.getBeamType() > 0)
            return makeStringValue(::getBeam(sh.getBeamType()).getName());
        else
            return 0;
     case ispCARGO_COLONISTS:
        /* colonists */
        if (sh.getCargoRaw(el_Colonists) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Colonists));
        else
            return 0;
     case ispCARGO_D:
        /* duranium */
        if (sh.getCargoRaw(el_Duranium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Duranium));
        else
            return 0;
     case ispCARGO_FREE:
        /* free cargo */
        return makeIntValue(sh.getFreeCargo());
     case ispCARGO_M:
        /* molybdenum */
        if (sh.getCargoRaw(el_Molybdenum) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Molybdenum));
        else
            return 0;
     case ispCARGO_MONEY:
        /* money */
        if (sh.getCargoRaw(el_Money) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Money));
        else
            return 0;
     case ispCARGO_N:
        /* fuel */
        if (sh.getCargoRaw(el_Neutronium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Neutronium));
        else
            return 0;
     case ispCARGO_STR:
        /* FIXME */
        return 0;
     case ispCARGO_SUPPLIES:
        /* supplies */
        if (sh.getCargoRaw(el_Supplies) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Supplies));
        else
            return 0;
     case ispCARGO_T:
        /* supplies */
        if (sh.getCargoRaw(el_Tritanium) >= 0)
            return makeIntValue(sh.getCargoRaw(el_Tritanium));
        else
            return 0;
     case ispCREW:
        /* crew */
        return makeIntFromMaybe(sh.getCrew());
     case ispDAMAGE:
        /* damage */
        return makeIntFromMaybe(sh.getDamage());
     case ispENGINE_:
        /* engine id */
        return makeIntFromMaybe(sh.getEngine());
     case ispENGINE:
        /* engine name */
        if (sh.getEngine().isKnown())
            return makeStringValue(::getEngine(sh.getEngine()).getName());
        else
            return 0;
     case ispFCODE:
        /* fcode */
        // FIXME:
        if (sh.getFCode().size())
            return makeStringValue(sh.getFCode());
        else
            return 0;
     case ispFIGHTER_BAYS:
        /* fighter bays */
        return makeIntFromMaybe(sh.getBayCount());
     case ispFIGHTER_COUNT:
        /* fighter count */
        if (sh.getBayCount().isKnown() && sh.getBayCount() > 0 && sh.getAmmo() >= 0)
            return makeIntValue(sh.getAmmo());
        else
            return 0;
//   FLEET$: INT                            OWN
//   FLEET.NAME: STR               R/W      OWN
//   FLEET.STATUS: STR                      OWN
//   FLEET: STR                             OWN
     case ispHEADING_:
        /* heading angle */
        return makeIntFromMaybe(sh.getHeading());
     case ispHEADING:
        /* heading name */
        // FIXME: this goes through NLS. Should it?
        if (sh.getHeading().isKnown())
            return makeStringValue(getHeadingName(sh.getHeading()));
        else
            return 0;
     case ispHULL_SPECIAL:
        /* Special flags */
        {
            string_t result;
            if (sh.isExistant()) {
                if (sh.canDoSpecial(hf_Cloak) || sh.canDoSpecial(hf_AdvancedCloak) || sh.canDoSpecial(hf_HardenedCloak))
                    result += "C";
                if (sh.canDoSpecial(hf_Hyperdrive))
                    result += "H";
                if (sh.canDoSpecial(hf_Gravitonic))
                    result += "G";
                if (sh.canDoSpecial(hf_Bioscan) || sh.canDoSpecial(hf_FullBioscan))
                    result += "B";
                if (sh.canDoSpecial(hf_MerlinAlchemy) || sh.canDoSpecial(hf_AriesRefinery) || sh.canDoSpecial(hf_NeutronicRefinery))
                    result += "A";
            }
            return makeStringValue(result);
        }
     case ispID:
        /* ship Id */
        return makeIntValue(sh.getId());
     case ispLEVEL:
        {
            int16_t level = sh.unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
            if (level >= 0)
                return makeIntValue(level);
            else
                return 0;
        }
     case ispLOC_X:
        /* ship location */
        if (sh.isExistant())
            return makeIntValue(sh.getPos().x);
        else
            return 0;
     case ispLOC_Y:
        /* ship location */
        if (sh.isExistant())
            return makeIntValue(sh.getPos().y);
        else
            return 0;
     case ispLOC:
        if (sh.isExistant())
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getLocationName(sh.getPos(), 0));
        else
            return 0;
     case ispMARKED:
        /* marked */
        return makeBoolValue(sh.isMarked());
     case ispMASS:
        /* mass */
        if (sh.isExistant())
            return makeIntValue(sh.getMass());
        else
            return 0;
     case ispMISSION_:
        /* mission */
        return makeIntFromMaybe(sh.getMission());
     case ispMISSION_INTERCEPT:
        /* Intercept Id */
        return makeIntFromMaybe(sh.getIntId());
//   MISSION.SHORT: STR                     OWN
     case ispMISSION_TOW:
        /* Tow Id */
        return makeIntFromMaybe(sh.getTowId());
     case ispMISSION:
        /* mission name */
        if (sh.getMission().isKnown()) {
            const GMission* m = GMissionList::getInstance().getMissionByNumber(sh.getMission(), GPlayerSet(sh.getRealOwner()));
            if (m != 0) {
                return makeStringValue(m->getName());
            } else {
                // FIXME: 1.x is completely different here
                return makeIntValue(sh.getMission());
            }
        } else {
            return 0;
        }
     case ispMOVE_ETA:
        if (sh.getKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(getDisplayedTurn().getCurrentUniverse(), sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getTurnCount());
        } else {
            return 0;
        }
     case ispMOVE_FUEL:
        if (sh.getKind() == sh.CurrentShip) {
            GShipTurnPredictor pred(getDisplayedTurn().getCurrentUniverse(), sh.getId());
            pred.computeMovement();
            return makeIntValue(pred.getFuelUsed());
        } else {
            return 0;
        }
     case ispNAME:
        if (sh.isExistant())
            return makeStringValue(sh.getName());
        else
            return 0;
     case ispORBIT_:
        if (sh.isExistant() && sh.getOrbit() > 0)
            return makeIntValue(sh.getOrbit());
        else
            return 0;
     case ispORBIT:
        if (sh.isExistant() && sh.getOrbit() > 0)
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sh.getOrbit()).getName());
        else
            return 0;
     case ispOWNER_REAL:
        return makeIntValue(sh.getRealOwner());
     case ispSPEED_:
        return makeIntFromMaybe(sh.getWarp());
     case ispSPEED:
        if (sh.getWarp().isKnown()) {
            if (sh.isHyperdriving())
                return makeStringValue("Hyperdrive");
            else
                return makeStringValue(format("Warp %d", sh.getWarp()));
        } else {
            return 0;
        }
//   TASK: BOOL                             ALWAYS
     case ispTORP_:
        return makeIntFromMaybe(sh.getTorpType());
     case ispTORP_COUNT:
        if (sh.getTorpType().isKnown()) {
            if (sh.getTorpType() > 0)
                return makeIntValue(sh.getAmmo());
            else
                return makeIntValue(0);
        } else {
            return 0;
        }
     case ispTORP_LCOUNT:
        return makeIntFromMaybe(sh.getTorpLauncherCount());
     case ispTORP_SHORT:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getShortName());
        else
            return 0;
     case ispTORP:
        if (sh.getTorpType().isKnown() && sh.getTorpType() > 0)
            return makeStringValue(::getTorp(sh.getTorpType()).getName());
        else
            return 0;
     case ispTRANSFER_SHIP_COLONISTS:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).colonists);
        else
            return 0;
     case ispTRANSFER_SHIP_D:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Duranium]);
        else
            return 0;
     case ispTRANSFER_SHIP_ID:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).id);
        else
            return 0;
     case ispTRANSFER_SHIP_M:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Molybdenum]);
        else
            return 0;
     case ispTRANSFER_SHIP_N:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Neutronium]);
        else
            return 0;
     case ispTRANSFER_SHIP_NAME:
        if (sh.getKind() == sh.CurrentShip
            && sh.isTransporterActive(sh.t_Transfer)
            && getDisplayedTurn().getCurrentUniverse().isValidShipId(sh.getTransporter(sh.t_Transfer).id))
        {
            return makeStringValue(getDisplayedTurn().getCurrentUniverse().getShip(sh.getTransporter(sh.t_Transfer).id).getName());
        } else {
            return 0;
        }
     case ispTRANSFER_SHIP_T:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).ore[el_Tritanium]);
        else
            return 0;
     case ispTRANSFER_SHIP_SUPPLIES:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Transfer).supplies);
        else
            return 0;
     case ispTRANSFER_SHIP:
        if (sh.getKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Transfer));
        else
            return 0;
     case ispTRANSFER_UNLOAD_COLONISTS:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).colonists);
        else
            return 0;
     case ispTRANSFER_UNLOAD_D:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Duranium]);
        else
            return 0;
     case ispTRANSFER_UNLOAD_ID:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).id);
        else
            return 0;
     case ispTRANSFER_UNLOAD_M:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Molybdenum]);
        else
            return 0;
     case ispTRANSFER_UNLOAD_N:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Neutronium]);
        else
            return 0;
     case ispTRANSFER_UNLOAD_NAME:
        if (sh.getKind() == sh.CurrentShip && sh.isTransporterActive(sh.t_Unload)) {
            if (sh.getTransporter(sh.t_Unload).id == 0)
                return makeStringValue("Jettison");
            else if (getDisplayedTurn().getCurrentUniverse().isValidPlanetId(sh.getTransporter(sh.t_Unload).id))
                return makeStringValue(getDisplayedTurn().getCurrentUniverse().getPlanet(sh.getTransporter(sh.t_Unload).id).getName());
            else
                return 0;
        } else {
            return 0;
        }
     case ispTRANSFER_UNLOAD_T:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).ore[el_Tritanium]);
        else
            return 0;
     case ispTRANSFER_UNLOAD_SUPPLIES:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getTransporter(sh.t_Unload).supplies);
        else
            return 0;
     case ispTRANSFER_UNLOAD:
        if (sh.getKind() == sh.CurrentShip)
            return makeBoolValue(sh.isTransporterActive(sh.t_Unload));
        else
            return 0;
     case ispTYPE_SHORT:
        if (const char* p = classifyShip(sh))
            return makeStringValue(string_t(p, 1));
        else
            return 0;
     case ispTYPE:
        if (const char* p = classifyShip(sh))
            return makeStringValue(p);
        else
            return 0;
     case ispWAYPOINT_DIST:
        if (sh.getKind() == sh.CurrentShip)
            return makeFloatValue(std::sqrt(double(square(sh.getWaypointDX()) + square(sh.getWaypointDY()))));
        else
            return 0;
     case ispWAYPOINT_DX:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDX());
        else
            return 0;
     case ispWAYPOINT_DY:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDY());
        else
            return 0;
     case ispWAYPOINT_PLANET:
        if (sh.getKind() == sh.CurrentShip) {
            return makeIntValue(getDisplayedTurn().getCurrentUniverse().getPlanetAt(sh.getWaypoint(), true));
        } else {
            return 0;
        }
     case ispWAYPOINT_X:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDX() + sh.getPos().x);
        else
            return 0;
     case ispWAYPOINT_Y:
        if (sh.getKind() == sh.CurrentShip)
            return makeIntValue(sh.getWaypointDY() + sh.getPos().y);
        else
            return 0;
     case ispWAYPOINT:
        if (sh.getKind() == sh.CurrentShip) {
            // FIXME: PCC 1.x also handles Intercept here
            if (sh.getWaypointDX() == 0 && sh.getWaypointDY() == 0)
                return makeStringValue("(Location)");
            else
                return makeStringValue(getDisplayedTurn().getCurrentUniverse().getLocationName(sh.getWaypoint().getNormalized(), 0));
        } else {
            return 0;
        }
     default:
        return 0;
    }
}


void
setShipProperty(GShip& sh, IntShipProperty isp, IntValue* value)
{
//   ENEMY$: INT                   R/W      OWN
//   FCODE: STR                    R/W      OWN
//   FLEET.NAME: STR               R/W      OWN
//   MISSION$: INT                 R/W      OWN
//   MISSION.INTERCEPT: INT        R/W      OWN
//   MISSION.TOW: INT              R/W      OWN
//   NAME: STR                     R/W      OWN
//   SPEED$: INT                   R/W      OWN
    throw IntError::notAssignable();
}
