/**
  *  \file int/value.h
  *  \brief Interpreter: Values
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_VALUE_H
#define PCC_V2_INT_VALUE_H

#include "cpluslib/string.h"
#include "cpluslib/types.h"

class Stream;
class TStructureParser;

/** Serialized form of a value. This is the same structure as used in PCC 1.x.
    A serialized data segment consists of a sequence of these 48-bit tag nodes,
    followed by (optional) auxiliary data. The tag nodes contain
    - 32-bit integer or bool
    - string tag (actual string data in auxiliary data field)
    - 48-bit floating point values (Turbo Pascal REAL)
    - 32-bit floating point values (IEEE single precision)

    Turbo Pascal has an entirely-software emulated floating point type REAL. Thoes
    have the convenient property of having an effective value of 0.0 if their first
    byte, the exponent, is zero. All the other bytes are irrelevant in this case
    (except for the 47th bit, the sign bit, which distinguishes positive and negative
    zeroes, to be precise). PCC 1.x therefore makes sure that actual floaring-point
    zeroes are stored as all-48-bits-zero, and uses the "non-standard" zeroes
    to store tag/value pairs. The tags thus need to have zero in the lower 8 bits.

    Non-scalar values require additional data in the auxiliary data field. This data
    follows the tag nodes in sequence. This has the drawback that random access is
    not possible, just sequential reading, and it also means that there is no way
    to skip over data added by potential future tags, i.e. non-scalars after those
    future values are lost. PCC 1.x, in anticipation of PCC II, implements a little
    forward compatibility by reading 32-bit floats and long strings, although it
    never writes them. */
struct IntTagNode {
    uint16 tag;
    uint32 value;

    enum {
        Tag_FPZero     = 0x0000,       ///< Floating-point zero.
        Tag_Empty      = 0x0100,       ///< EMPTY (null, unassigned, ...).
        Tag_Integer    = 0x0200,       ///< Integer. value is actual value.
        Tag_Boolean    = 0x0300,       ///< Boolean. value is 0 or 1.
        Tag_String     = 0x0400,       ///< String. value is flag: 0=empty string, 1=Pascal string follows in auxiliary data.
        Tag_32bitFP    = 0x0500,       ///< 32-bit floating point. value is IEEE single precision value. Supported since PCC 1.0.9.
        Tag_LongString = 0x0600        ///< String. value is number of bytes following in auxiliary data. Supported since PCC 1.0.18.
    };

    // Serialisation information
    static const TStructureParser parser[];
    enum {
        size  = 6,
        count = 2
    };
};

/** Generic value. This is the base class for all values used in the interpreter.
    Note that all places that use an IntValue* also allow it to be null, representing
    the CCScript EMPTY (null) value used for unknown properties. */
class IntValue {
 public:
    /** Destructor. */
    virtual ~IntValue();

    /** Convert this value to a string.
        \param readable true=make it readable to the lexer; false=program-controlled interpolation */
    virtual string_t toString(bool readable) = 0;

    /** Store this value to a stream.
        \param sv [out] Header (tag) information goes here
        \param aux [out] Auxiliary data stored here, if required
        \see IntTagNode */
    virtual void store(IntTagNode& sv, Stream& aux) = 0;

    /** Clone this value. Returns a duplicate of this value. The caller assumes responsibility for
        the new object. */
    virtual IntValue* clone() = 0;

    /** Load a value.
        \param sv  [in] Header (tag) information
        \param aux [in] Auxiliary data stream, auxiliary data read here if required
        \return newly-allocated IntValue descendant (or null)
        \see IntTagNode */
    static IntValue* load(const IntTagNode& sv, Stream& aux);

    /** Clone a value. Same as p->clone(), but allows p to be 0. */
    static IntValue* cloneOf(IntValue* p);

    /** Convert a value into a string. Same as p->toString(readable), but allows p to be 0. */
    static string_t toStringOf(IntValue* p, bool readable);
};

/** Integral value. This represents an integer or boolean value. Those behave identical
    in all operations, but we track whether it's an integer or boolean to support meaningful
    stringification. */
class IntIntValue : public IntValue {
    int32 value;
    bool is_bool;
 public:
    IntIntValue(int32 value, bool is_bool);
    IntIntValue(const IntTagNode& sv, Stream& aux);
    ~IntIntValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntIntValue* clone();

    /** Get integer value. */
    int32 getValue() const
        { return value; }
    /** Check whether this is a boolean value. */
    bool isBool() const
        { return is_bool; }
};

/** String value. */
class IntStringValue : public IntValue {
    string_t value;
 public:
    IntStringValue(string_t value);
    IntStringValue(const IntTagNode& sv, Stream& aux);
    ~IntStringValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntStringValue* clone();

    /** Get string value. */
    string_t getValue() const
        { return value; }
};

/** Floating-Point Value. We represent floating point as double internally,
    to achieve some mentionable precision. For external storage, floating-point
    values are converted to 48-bit format (the same as used in PCC 1.x). */
class IntFloatValue : public IntValue {
    double value;
 public:
    IntFloatValue(double value);
    IntFloatValue(const IntTagNode& sv, Stream& aux);
    ~IntFloatValue();

    string_t toString(bool readable);
    void     store(IntTagNode& sv, Stream& aux);
    IntFloatValue* clone();

    double getValue() const
        { return value; }
};

IntValue* makeBoolValue(int value);
IntValue* makeIntValue(int32_t value);
IntValue* makeFloatValue(double value);
IntValue* makeStringValue(string_t str);
IntValue* makeStringValue(const char* str);
int       getBoolValue(IntValue* value);

string_t  quoteString(const string_t& value);

#endif
