/**
  *  \file ui/widgets.cc
  *  \brief Standard Widgets
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "ui/widgets.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/layout.h"
#include "ui/skin.h"

static inline char lc(char c) {
    using namespace std;
    return tolower((unsigned char) c);
}

UIKeyString::UIKeyString(const char* s)
    : k(lc(s[0])), s(s)
{ }

UIKeyString::UIKeyString(const string_t& s)
    : k(lc(s[0])), s(s)
{ }

UIKeyString::~UIKeyString()
{ }

/***************************** UIImageWidget *****************************/

/** \class UIImageWidget
    \brief Static widget displaying a pixmap */
UIImageWidget::UIImageWidget(Ptr<GfxPixmap> apix)
    : UIWidget(0), pix(apix),
      size(pix->getWidth(), pix->getHeight())
{ }

UIImageWidget::UIImageWidget(int x, int y, Ptr<GfxPixmap> pix)
    : UIWidget(0), pix(pix),
      size(x, y)
{ }

UIImageWidget::UIImageWidget(int x, int y)
    : UIWidget(0), pix(0),
      size(x, y)
{ }

void
UIImageWidget::drawContent(GfxCanvas& can)
{
    can.drawBar(extent, standard_colors[0], GFX_TRANSPARENT, SOLID_FILL, GFX_ALPHA_OPAQUE);
    drawFrameDown(can, extent);
    if (pix) {
        GfxRect clip(0, 0,
                     std::min(extent.w - 2, pix->getWidth()),
                     std::min(extent.h - 2, pix->getHeight()));
        can.blitSurface(GfxPoint(extent.x+1, extent.y+1), *pix, clip);
    }
    can.update(extent);
}

void
UIImageWidget::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = size;
    info.flags = 0;
}

void
UIImageWidget::setImage(Ptr<GfxPixmap> pix)
{
    this->pix = pix;
    drawWidget(false);
}

/****************************** UIStaticText *****************************/

/** \class UIStaticText
    \brief Static line of text

    This widget can display a particular piece of text in one of several
    styles. */

static const int text_types[] = {
    FONT_TITLE,
    FONT_NORMAL,
    FONT_TITLE,
    FONT_TITLE,
    FONT_NORMAL
};
static const int text_colors[] = {
    COLOR_BLACK,
    COLOR_BLACK,
    COLOR_GREENSCALE + 7,
    COLOR_WHITE,
    COLOR_WHITE
};

UIStaticText::UIStaticText(const string_t& text, Scheme scheme, int align)
    : UIWidget(0), text(text), scheme(scheme), align(align)
{
    setState(st_Disabled, true);
}

UIStaticText::UIStaticText(const char* text, Scheme scheme, int align)
    : UIWidget(0), text(text), scheme(scheme), align(align)
{
    setState(st_Disabled, true);
}

void
UIStaticText::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size =
        GfxPoint(fonts[text_types[scheme]]->getTextWidth(text),
                 font_heights[text_types[scheme]]);
}

void
UIStaticText::drawContent(GfxCanvas& can)
{
    // FIXME: convert this to use the skin, not just the scheme
    GfxParameters parm;
    parm.color = standard_colors[text_colors[scheme]];
    parm.font = fonts[text_types[scheme]];
    getSkin().drawBackground(can, getExtent());

    int dx = 0;
    if (align != 0)
        dx = (extent.w - parm.font->getTextWidth(text))*align / 2;

    outTextF(can, parm, extent.x + dx, extent.y, extent.w - dx, text);
    can.update(extent);
}

void
UIStaticText::setText(const string_t& text)
{
    this->text = text;
    drawWidget();
}

/**************************** UIFocusIterator ****************************/

/** \class UIFocusIterator
    \brief Cursor Focus Control

    This invisible widget provides a means of manipulating focus using
    the keyboard. Given a range of widget Ids, it provides keys to
    move within this range.
    - fi_Horiz: left and right arrows
    - fi_Vert: up and down arrows
    - fi_Tab: Tab and Shift-Tab keys
    - fi_Home: accepts Home and End to move to ends
    - fi_Page: accepts PgUp and PgDn to move to ends
    - fi_Wrap: when added to fi_Horiz/fi_Vert, those "wrap around". By
      default, only fi_Tab wraps around. */

UIFocusIterator::UIFocusIterator(int from, int to, int flags)
    : UIFixedWidget(), from(from), to(to), flags(flags)
{
    setState(st_Disabled, true);  // so it doesn't try to get the focus
}

UIFocusIterator::~UIFocusIterator()
{ }

void
UIFocusIterator::drawContent(GfxCanvas& /*can*/)
{ }

bool
UIFocusIterator::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        int type = 0;
        switch(e.key.code) {
         case SDLK_TAB:            type = fi_Tab; break;
         case SDLK_TAB + ss_Shift: type = fi_Tab + fi_INTERNAL_NEGATE; break;
         case SDLK_UP:             type = fi_Vert + fi_INTERNAL_NEGATE; break;
         case SDLK_DOWN:           type = fi_Vert; break;
         case SDLK_LEFT:           type = fi_Horiz + fi_INTERNAL_NEGATE; break;
         case SDLK_RIGHT:          type = fi_Horiz; break;
         case SDLK_HOME:
         case SDLK_HOME + ss_Ctrl: type = fi_Home + fi_INTERNAL_NEGATE; break;
         case SDLK_END:
         case SDLK_END + ss_Ctrl:  type = fi_Home; break;
         case SDLK_PAGEUP:         type = fi_Page + fi_INTERNAL_NEGATE; break;
         case SDLK_PAGEDOWN:       type = fi_Page; break;
        }
        if (type & flags) {
            /* Locate a focused widget which we know. Do not go past our parent. */
            UIBaseWidget* focused_w = getParent()->getFocusedWidget();
            while (focused_w != 0
                   && (focused_w->getId() < from || focused_w->getId() > to)
                   && focused_w != getParent())
            {
                focused_w = focused_w->getParent();
            }

            type |= flags & fi_Wrap;

            /* If we found one, we know the focused Id. Find new Id */
            if (focused_w != 0 && focused_w != getParent()) {
                int focus_id = focused_w->getId();
                if (type & (fi_Page | fi_Home)) {
                    /* in Page/Home mode, move backward from end */
                    if (type & fi_INTERNAL_NEGATE) {
                        focus_id = from-1;
                        type = 0;
                    } else {
                        focus_id = to+1;
                        type = fi_INTERNAL_NEGATE;
                    }
                }

                int new_id = focus_id;
                UIBaseWidget* new_widget = 0;
                do {
                    if (type & fi_INTERNAL_NEGATE)
                        --new_id;
                    else
                        ++new_id;
                    if (new_id < from) {
                        if (type & (fi_Tab | fi_Wrap))
                            new_id = to;
                        else
                            return true;
                    } else if (new_id > to) {
                        if (type & (fi_Tab | fi_Wrap))
                            new_id = from;
                        else
                            return true;
                    }
                    new_widget = getParent()->getWidgetById(new_id);
                } while (new_id != focus_id && (new_widget == 0 || new_widget->hasState(st_Disabled)));
                if (new_id != focus_id) {
                    setSelectedWidget();
                    new_widget->setFocusedWidget();
                }
                return true;
            }
        } /* not a focus iterator key */
    } /* not a key */
    return UIFixedWidget::handleEvent(e, second_pass);
}

/****************************** UIFrameGroup *****************************/

/** \class UIFrameGroup
    \brief Framed Group

    This widget is like UIGroup, but it will draw a "down" frame
    around its child widgets. */

UIFrameGroup::UIFrameGroup(UILayoutManager& lay, int frame)
    : UILayoutableWidget(lay), frame(frame)
{ }

void
UIFrameGroup::drawContent(GfxCanvas& can)
{
    drawFrameDown(can, getExtent());
    can.update(getExtent());
}

void
UIFrameGroup::adjustSize(GfxRect& size, bool to_inner)
{
    if (to_inner)
        size.grow(-frame, -frame);
    else
        size.grow(frame, frame);
}

/****************************** UIFocusGroup *****************************/

/** \class UIFocusGroup
    \brief Group that can have Focus, Visibly

    This is a group which can indicate visibly that it has focus.
    It also ensures that keyboard focus is distributed correctly.
    Otherwise it is exactly like UIGroup.

    \todo Handling of disabled groups is not yet complete.
    - we should implement getCanvasFor() to allow children's
      drawWidget(true) requests. This would require implementation
      of a canvas that performs automatic blending upon update().
      It would make sense to use that in drawEverything() as well.
    - support for drawWidget(false) (see onStateChange()) is a hack */
UIFocusGroup::UIFocusGroup(UILayoutManager& lay, int id, int frame)
    : UILayoutableWidget(lay, id), frame(frame)
{ }

UIFocusGroup::~UIFocusGroup()
{ }

void
UIFocusGroup::drawContent(GfxCanvas& can)
{
    UISkin& s = getSkin();
    if (getFocusState() == NoFocus) {
        s.drawBackground(can, GfxRect(extent.x, extent.y, extent.w, 1));
        s.drawBackground(can, GfxRect(extent.x, extent.y+extent.h-1, extent.w, 1));
        s.drawBackground(can, GfxRect(extent.x, extent.y+1, 1, extent.h-2));
        s.drawBackground(can, GfxRect(extent.x+extent.w-1, extent.y+1, 1, extent.h-2));
    } else {
        s.setColor(UIColor::tc_Contrast);
        drawRectangle(can, s, extent);
    }
    can.update(extent);
}

void
UIFocusGroup::drawEverything(GfxCanvas& can)
{
    UILayoutableWidget::drawEverything(can);
    if (hasState(st_Disabled)) {
        uint32 bgcolor = standard_colors[getSkin().getColors()[UIColor::tc_Background]];
        if (ui_screen_bits < 16) {
            can.drawBar(getExtent(), bgcolor, GFX_TRANSPARENT, GRAY50_FILL, GFX_ALPHA_OPAQUE);
        } else {
            can.drawBar(getExtent(), bgcolor, GFX_TRANSPARENT, SOLID_FILL, 192);
        }
        can.update(getExtent());
    }
}

void
UIFocusGroup::onStateChange(int state, bool enable)
{
    if (state & (st_Focused | st_Disabled))
        drawWidget();

    /* If a child wants to be redrawn, redraw this group completely to make sure
       the "dimming" feature correctly propagates. */
    if (enable && (state & st_ChildRedraw) && hasState(st_Disabled))
        this->status |= st_WantRedraw;

    UILayoutableWidget::onStateChange(state, enable);
}

bool
UIFocusGroup::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Disabled))
        return false;
    if (e.type == ev_Key && !hasState(st_Focused))
        return false;
    if (e.type == ev_Mouse && e.mouse.b
        && !hasState(st_Focused) && isInside(GfxPoint(e.mouse.x, e.mouse.y)))
    {
        setSelectedWidget();
        setFocusedWidget();
        /* FALLTHROUGH */
    }
    return UILayoutableWidget::handleEvent(e, second_pass);
}

void
UIFocusGroup::adjustSize(GfxRect& size, bool to_inner)
{
    if (to_inner)
        size.grow(-frame, -frame);
    else
        size.grow(frame, frame);
}
